# -*- coding: utf-8 -*-
"""
    Modul pro práci s DCEL.
"""

from __future__ import unicode_literals

__author__ = 'Dominik Janků'
__license__ = 'BSD'
__version__ = '1.0'


import os
import xml.dom.minidom

from geotools import Point, Segment


class Vertex(object):
    def __init__(self, name, point, edge=None):
        self.name = name
        self.point = point
        self.edge = edge

    def __str__(self):
        return "Vertex({0}, [{1},{2}], {3})".format(self.name, self.point.x, self.point.y,
                                                    self.edge.name)

    def __repr__(self):
        return "<Vertex({0}, [{1},{2}], {3})>".format(self.name, self.point.x, self.point.y,
                                                      self.edge.name)

    @property
    def xml(self):
        return '  <vertex id="{0}" x="{1.x}" y="{1.y}" edge="{2}" />\n'.format(
            self.name, self.point, self.edge.name)


class Edge(object):
    def __init__(self, name, origin=None, twin=None, next=None, prev=None, face=None, segment=None):
        self.name = name
        self.origin = origin
        self.twin = twin
        self.next = next
        self.prev = prev
        self.face = face

        # pomocna - ukazatel na usecku
        self.segment = segment

    def __str__(self):
        return "Edge(%s, %s, twin=%s, prev=%s, next=%s, face=%s)" % \
               (self.name, self.origin.name, self.twin.name, self.prev.name,
               self.next.name, self.face.name if self.face is not None else "-")

    @property
    def xml(self):
        return '  <edge id="{0.name}" origin="{0.origin.name}" ' \
               'twin="{0.twin.name}" next="{0.next.name}" prev="{0.prev.name}" ' \
               'face="{0.face.name}" />\n'.format(self)


class Face(object):
    def __init__(self, name, outer_edge=None, inner_edges=[]):
        self.name = name
        self.outer_edge = outer_edge
        self.inner_edges = inner_edges

    def __str__(self):
        outer = 'neni' if self.outer_edge is None else self.outer_edge.name
        inners = '[]' if not len(self.inner_edges) else \
            ", ".join(map(lambda x: x.name, self.inner_edges))

        return "Face(%s, %s, inners=%s)" % (self.name, outer, inners)

    @property
    def xml(self):
        data = '  <face id="{0}">\n'.format(self.name)

        if self.outer_edge is not None:
            data += '    <border type="outer" edge="{0}" />\n'.format(self.outer_edge.name)

        for inner in self.inner_edges:
            data += '    <border type="inner" edge="{0}" />\n'.format(inner.name)

        data += '  </face>\n'
        return data



class DCEL(object):
    def __init__(self, name=None):
        self.name = name
        self.vertices = []
        self.edges = []
        self.faces = []

        # pomocna promenna (pro ucely gui)
        self.segments = []

        # vnejsi neomezena oblast (muze se hodit)
        self.outer_face = None

    def print_dcel(self):
        print self.name
        print "-"*len(self.name)
        print "Vrcholy:"
        for v in self.vertices:
            print v
        print ""
        print "Hrany:"
        for e in self.edges:
            print e
        print ""
        print "Oblasti:"
        for f in self.faces:
            print f
        print ""

    def load_from_file(self, filename):
        """
        Nahraje dvojitě souvislý seznam z XML souboru.
        """
        if not os.path.exists(filename):
            print "Soubor", filename, "neexistuje."
            return None

        _vertices = {}
        _edges = {}
        _faces = {}

        dom = xml.dom.minidom.parse(filename)

        # vezmu si prvni (nultou mapu v XML)
        dom_map = dom.getElementsByTagName('map')[0]
        _name = dom_map.attributes['name'].value

        try:
            dom_vertices = dom_map.getElementsByTagName('vertices')[0]
            # nahraji data vrcholu
            for vertex in dom_vertices.getElementsByTagName('vertex'):
                name = vertex.attributes['id'].value
                x = int(float(vertex.attributes['x'].value))
                y = int(float(vertex.attributes['y'].value))
                edge = vertex.attributes['edge'].value

                _vertices[name] = Vertex(name, Point(x, y), edge)

            dom_edges = dom_map.getElementsByTagName('edges')[0]
            # nahraji data vrcholu
            for edge in dom_edges.getElementsByTagName('edge'):
                name = edge.attributes['id'].value
                origin = edge.attributes['origin'].value
                next = edge.attributes['next'].value
                prev = edge.attributes['prev'].value
                twin = edge.attributes['twin'].value
                face = edge.attributes['face'].value

                _edges[name] = Edge(name, origin, twin, next, prev, face)

            dom_faces = dom_map.getElementsByTagName('faces')[0]
            # nahraji data oblasti
            for face in dom_faces.getElementsByTagName('face'):
                name = face.attributes['id'].value
                outer = None
                inners = []

                # zjistim hranice oblasti
                for border in face.getElementsByTagName('border'):
                    type = border.attributes['type'].value
                    edge = border.attributes['edge'].value
                    if type == 'inner':
                        inners.append(_edges[edge])
                    elif type == 'outer':
                        outer = _edges[edge]
                    else:
                        raise ValueError("Nespravny typ hranice u oblasti {0}. "
                                         "Mozny je jen 'inner' a 'outer'.".format(name))

                face = Face(name, outer, inners)

                if outer is None:
                    self.outer_face = face

                _faces[name] = face

            # predelame nazvy objektu na odkazy na ne a zacneme vrcholy
            for vertex in _vertices.itervalues():
                vertex.edge = _edges[vertex.edge]

            # predelame nazvy objektu na odkazy u hran
            for edge in _edges.itervalues():
                edge.next = _edges[edge.next]
                edge.prev = _edges[edge.prev]
                edge.twin = _edges[edge.twin]
                edge.origin = _vertices[edge.origin]
                edge.face = _faces[edge.face]

        except KeyError as err:
            print "Chyba pri nacitani XML"
            return False

        self.vertices = _vertices.values()
        self.edges = _edges.values()
        self.faces = _faces.values()
        self.name = _name
        self.update_segments()

    def clear(self):
        """
        Zruší všechna data z dvojitě souvislého seznamu.
        """
        self.edges = []
        self.vertices = []
        self.faces = []

    def save_to_file(self, filename):
        pass

    def update_segments(self):
        """
        Nastavi seznam úseček, který je sestaven z hran.
        """
        self.segments = []

        # Provedeme prevod hran na usecky (pro ucely pocitani pruseciku)
        map(lambda e: setattr(e, 'is_processed', False), self.edges)

        for e in self.edges:
            if not e.is_processed:
                name = e.name + e.twin.name
                segment = Segment(e.origin.point, e.twin.origin.point, name=name, edge=e)
                # nastavim odkaz hran na usecku
                e.segment = segment
                e.twin.segment = segment
                # dvojce je zpracovano
                e.twin.is_processed = True
                self.segments.append(segment)
        map(lambda e: delattr(e, 'is_processed'), self.edges)

    @property
    def xml(self):
        data = '<?xml version="1.0" encoding="UTF-8"?>\n'
        data += '<map name="export">\n'

        data += '<vertices>\n'
        for vertex in self.vertices:
            data += vertex.xml
        data += '</vertices>\n'

        data += '<edges>\n'
        for edge in self.edges:
            data += edge.xml
        data += '</edges>\n'

        data += '<faces>\n'
        for face in self.faces:
            data += face.xml
        data += '</faces>\n'

        data += '</map>\n'
        return data


