#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
"""
Program Sweepline (zametaci primka)
"""

from __future__ import unicode_literals

__author__ = 'Dominik Janků'
__license__ = 'BSD'
__version__ = '1.0'

from collections import defaultdict, namedtuple
from math import sqrt
from decimal import *
from Tkinter import *

from blist import sortedset, sortedlist
from geotools import Segment, Point

# preciznost na 8 desetinnych mist
getcontext().prec = 8


class SweepLineAlgorithm(object):
    """
    Trida reprezentujici algoritmus zametaci primky.
    """

    SearchSegClass = namedtuple('SearchSeg', ['a', 'b', 'name'])

    def __init__(self, canvas=None):
        # promenne simulatoru + algoritmu
        self.canvas = canvas

        #jednotlive kroky algoritmu
        self.steps = [
            ('input', 'Zadání úseček'),
            ('init', 'Inicializace'),
            ('event', 'Obsluha události p'),
            ('event-sets', 'zpracování U(p), C(p), L(p)'),
            ('event-update', 'aktualizace stromu'),
            ('event-findnew', 'hledání nových průsečíků'),
            ('end', 'Konec algoritmu', True),
        ]

        self.step_codes = map(lambda s: s[0], self.steps)
        self.steps_count = len(self.steps)

        # index kroku
        self.step = None
        self.next_step = 'input'

        self.can_step = True

        # text zobrazovany u daneho kroku
        self.step_info = ''

        # promenne algoritmu
        self.events = sortedset()
        self.event = None

        # pomocna promenna
        self._index = 0

        # jednotlive usecky {canvas_item: Segment}
        self.segments = list()

        # prusecik: mnozina usecek
        self.intersections = dict()

        self.Up = defaultdict(sortedset)
        self.Cp = defaultdict(sortedset)
        self.Lp = defaultdict(sortedset)

        # udalost, podle ktere se bude urcovat poradi usecek v T
        self._tree_event = None
        self.tree = sortedlist(key=self.tree_segment_order)

        # aktualni mnoziny
        self._up = None
        self._cp = None
        self._lp = None

        self._ucp_leftmost = None
        self._ucp_rightmost = None

    def add_segment(self, segment):
        """
        Přidá úsečku na vstup algoritmu.

        :param segment: Přidávaná úsečka.
        :type segment: Segment
        :return: Číslo úsečky.
        """
        self.segments.append(segment)
        segment.name = 's{0}'.format(len(self.segments)-1)
        return len(self.segments)-1

    def info_event(self):
        return "Událost p=[%.1f;%.1f]\n\n" % (self.event.x, self.event.y)

    def info_sets(self):
        return "L(p)={" + ", ".join([x.name for x in self._lp]) + "}\n" + \
               "C(p)={" + ", ".join([x.name for x in self._cp]) + "}\n" + \
               "U(p)={" + ", ".join([x.name for x in self._up]) + "}\n"

    def info_tree(self):
        text = "\nPořadí úseček ve stromě T:\n"
        if len(self.tree):
            text += ", ".join([s.name for s in self.tree]) + '\n'
        else:
            text += 'žádná úsečka\n'
        return text

    def tree_segment_order(self, s):
        """
        Urcuje poradi usecek ve strome T.
        :type s: Segment
        """
        p = self._tree_event

        a = min(s.a, s.b)
        b = max(s.a, s.b)

        x = Decimal(str(float(s.get_x_by_y(p.y)))[:-1]) if a.y != b.y else p.x
        try:
            alfa = (b.x-a.x) / sqrt((b.x-a.x)**2 + (b.y-a.y)**2)
        except ZeroDivisionError:
            alfa = 0

        return x, alfa

    def find_new_event(self, s1, s2):
        """
        Pokusi se nalezt novou udalost u dvou usecek.
        :type s1: Segment
        :type s2: Segment
        """
        if s1 is None or s2 is None:
            return None

        # otestuju dve usecky na jejich prusecik
        p = s1.has_intersection(s2)
        if p is None:
            return None

        # nova udalost lezi "nad" bodem stavajici udalosti => nezajima nas
        if p <= self.event:
            return None

        if p not in (s1.a, s1.b):
            self.Cp[p].add(s1)
        if p not in (s2.a, s2.b):
            self.Cp[p].add(s2)

        # ano, zjistena nova udalost
        if p not in self.events:
            self.events.add(p)
            return p
        return None

    def do_step(self):
        """
        Provede krok, ktery je definovany jako dalsi, tj. self.next_step.
        """
        self.step = self.next_step
        self.step_info = ''
        #print self.step, self.next_step
        handler = 'do_step_'+self.step.replace('-', '_')
        if hasattr(self, handler):
            f = getattr(self, handler)
            return f()

    def do_step_input(self):
        self.next_step = 'init'
        self.step_info = "Zadejte prosím vstupní úsečky pomocí myší."

    def do_step_init(self):
        if not len(self.segments):
            self.next_step = 'end'
            return []

        segment = self.segments[self._index]
        u, l = segment.upper_point, segment.lower_point

        self.Lp[l].add(segment)
        self.Up[u].add(segment)

        result = []
        u not in self.events and result.append(u)
        l not in self.events and result.append(l)

        # pridam body do seznamu udalosti
        map(lambda p: self.events.add(p), (u, l))

        if self._index+1 == len(self.segments):
            self.next_step = 'event'
            self._index = 0
        else:
            self.next_step = 'init'

        self.step_info = "Počet událostí: " + str(len(self.events)) + '\n'

        self._index += 1
        return result

    def do_step_event(self):
        if len(self.events):
            self.next_step = 'event-sets'
            self.event = self.events.pop(0)
            self.step_info = self.info_event()
            self.step_info += "Zbývá událostí: %d" % len(self.events)
            self.step_info += "\n\nOblast nad zametací přímkou a vlevo od bodu události má vyřešeny průsečíky."
            return self.event
        else:
            self.next_step = 'end'
            self.step_info = 'Žádná událost k obsluze.'
            return None

    def do_step_event_sets(self):
        self.next_step = 'event-update'
        self.step_info = self.info_event()

        self._lp = self.Lp[self.event]
        self._cp = self.Cp[self.event]
        self._up = self.Up[self.event]

        union = self._lp.union(self._cp, self._up)

        self.step_info += self.info_sets()

        self.step_info += self.info_tree()

        if len(union) > 1:
            self.step_info += "\n" + 'Detekován průsečík!'
            self.intersections[self.event] = union

        return union

    def do_step_event_update(self):
        """
        Aktualizace stromu T
        """
        self.next_step = 'event-findnew'
        self.step_info = self.info_event()

        if self._tree_event is None:
            self._tree_event = self.event

        # zrusim usecky L(p) ze stromu T (podle predchozi udalosti,
        # podle ktere se tam taky vkladaly)
        re_add = [s for s in self.tree if s not in self._lp.union(self._cp)]

        #print "STROM PRED: ", [s.name for s in self.tree]

        self._tree_event = self.event

        # tohle je smutnou nutnosti, protoze knihovna blist nevola key pri kazdem porovnavani,
        # ale patrne si pamatuje poradi interne, proto je treba to tam vsechno znovu pridat
        # dle stavajiciho bodu udalosti. Tim ovsem pada O(log n) ale jako demo to staci.
        new_tree = sortedlist(key=self.tree_segment_order)

        map(lambda s: new_tree.add(s), self._up.union(self._cp))
        map(lambda s: new_tree.add(s), re_add)

        #print "STROM PO: ", [s.name for s in new_tree]

        del self.tree
        self.tree = new_tree

        self.step_info += self.info_sets()
        self.step_info += self.info_tree()

    def do_step_event_findnew(self):
        self.step_info = self.info_event()
        self.step_info += self.info_sets()
        self.step_info += self.info_tree()

        new_events = []

        if len(self._up) + len(self._cp) == 0:
            search = self.SearchSegClass(a=self.event, b=self.event, name='search')
            i_l = self.tree.bisect_left(search)-1
            i_r = self.tree.bisect_right(search)
            sl = None if i_l < 0 else self.tree[i_l]
            sr = None if i_r > len(self.tree)-1 else self.tree[i_r]

            self.step_info += '\n' + "sl: {0}\nsr: {1}".format('není' if sl is None else sl.name,
                                                               'není' if sr is None else sr.name)
            new_events.append(self.find_new_event(sl, sr))
        else:
            ucp = self._up.union(self._cp)

            self.step_info += '\n'

            s_ll = min(ucp, key=self.tree_segment_order)
            self.step_info += "s': " + s_ll.name + '\nsl: '
            i = self.tree.index(s_ll)-1

            if i >= 0:
                s_l = self.tree[i]
                self.step_info += s_l.name
                new_events.append(self.find_new_event(s_ll, s_l))
            else:
                self.step_info += 'není'

            self.step_info += '\n\n'

            s_rr = max(ucp, key=self.tree_segment_order)
            self.step_info += "s'': " + s_rr.name + '\nsr: '
            i = self.tree.index(s_rr)+1

            if len(self.tree) > i:
                s_r = self.tree[i]
                self.step_info += s_r.name
                new_events.append(self.find_new_event(s_rr, s_r))
            else:
                self.step_info += 'není'

        self.next_step = 'event'

        return filter(lambda x: x is not None, new_events)

    def do_step_end(self):
        self.can_step = False
        self.step_info = "Nalezených průsečíků: {0}".format(len(self.intersections))


def vertical_separator(parent, row=0, column=0, rowspan=1, columnspan=1, **kwargs):
    """
    Vertikální oddělovač v GUI.
    """
    sep = Frame(parent, relief=RIDGE, width=2, bd=2, **kwargs)
    sep.grid(row=row, column=column, rowspan=rowspan, columnspan=columnspan, sticky=NSEW, padx=1)


def horizontal_separator(parent, row=0, column=0, rowspan=1, columnspan=1, **kwargs):
    """
    Horizontální oddělovač v GUI.
    """
    sep = Frame(parent, relief=RIDGE, height=2, bd=2, **kwargs)
    sep.grid(row=row, column=column, rowspan=rowspan, columnspan=columnspan, sticky=NSEW, pady=1)


class SweepLineGUI(object):
    FONT_STEP = "Font 13"
    FONT_SUB_STEP = "Font 12"

    def __init__(self, root):
        """
         :param root: Hlavní (mateřské okno)
         :type root: Tk
        """
        self._new_segment = None

        self.root = root
        root.title("Sweepline")
        root.minsize(900, 550)
        root.geometry("900x550+300+150")

        menu = Menu(root)

        # menu soubor
        filemenu = Menu(menu, tearoff=0)
        filemenu.add_command(label="Ukončit program", command=root.quit)
        menu.add_cascade(label="Soubor", menu=filemenu)

        root.config(menu=menu)

        # sestveni GUI
        self.canvas = Canvas(self.root, background='black', bd=0, cursor='crosshair')
        self.canvas.grid(row=0, column=0, sticky=NSEW, columnspan=2, pady=0, padx=0)

        self.canvas.bind('<Motion>', self.on_canvas_motion)
        self.canvas.bind('<Enter>', self.on_canvas_enter)
        self.canvas.bind('<Leave>', self.on_canvas_leave)
        self.canvas.bind('<Button-1>', self.on_canvas_btn1_click)
        self.canvas.bind('<B1-Motion>', self.on_canvas_btn1_motion)
        self.canvas.bind('<ButtonRelease-1>', self.on_canvas_btn1_release)

        # popisky jednotlivych objektu na platne (muzou byt prazdne)
        self.canvas_labels = dict()

        # uschova nejblizsi polozky (pro restore konfigurace)
        self._item_closest = None
        self._item_closest_config = None

        # punticek udalosti ke smazani
        self._event_to_delete = None

        # Algoritmus SWEEP LINE
        self.alg = SweepLineAlgorithm()

        # info o ukazateli mysi ([34;32])
        self.status_pos = Label(self.root, text="", width=20, bg='black', fg='white', font="Font 11")
        self.status_pos.grid(row=1, column=0, padx=3)

        self.status_item = Label(self.root, text="", bg='black', fg='white', anchor=W, padx=4,
                                 font='Font 11')
        self.status_item.grid(row=1, column=1, sticky=EW, padx=3)

        # vertikalni separator oddelujici canvas a info panel
        vertical_separator(self.root, row=0, column=2, rowspan=2)

        # --------------------- INFO -------------------------
        info = Frame(self.root, relief=FLAT, width=200, padx=4, pady=4)
        info.grid(column=3, row=0, rowspan=2, sticky=NSEW)

        l1 = Label(info, text="Kroky algoritmu:", anchor=W)
        l1.grid(row=0, column=0, sticky=EW)

        horizontal_separator(info, row=1, column=0)

        # ------------ INFO, FAZE ALGORITMU -------------------
        wsteps = Frame(info, pady=10)

        self.labels_steps = []

        for step in range(len(self.alg.steps)):
            code = self.alg.steps[step][0]
            text = self.alg.steps[step][1]
            if '-' in code:
                text = ' - ' + text
            label = Label(wsteps, text=text, anchor=W, padx=5)
            label.grid(row=step, sticky=W)
            # naformatovani fontu a barvy probiha ve funkci self.update_steps
            self.labels_steps.append(label)

        wsteps.grid(row=2, sticky=EW)

        horizontal_separator(info, row=3, pady=10)

        # ------------ INFO ovladaci prvky
        pw = PanedWindow(info, orient=HORIZONTAL)
        pw.grid(row=4, sticky=EW)

        self.btn_step = Button(pw, text="Dále", command=self.on_step_click, width=10)
        pw.add(self.btn_step, sticky=EW)

        self.btn_next = Button(info, text="Vše", command=self.on_all_click)
        pw.add(self.btn_next, sticky=EW)

        horizontal_separator(info, row=5)

        # ------------ INFO zobrazovani info k aktualnimu kroku
        self.step_info = Label(info, text='', anchor=W, justify=LEFT, wraplength=195, font="Font 12")
        self.step_info.grid(row=6, sticky=NSEW, pady=5)


        # ----------- Tlacitko na ukonceni aplikace
        pw2 = PanedWindow(info, orient=VERTICAL)
        pw2.add(Frame(pw2, relief=RIDGE, height=2, pady=4, bd=2))
        # Tlacitko na ukoneceni aplikace
        self.btn_quit = Button(info, text="Ukončit program", fg='red', command=self.root.destroy)
        pw2.add(self.btn_quit)
        pw2.grid(row=20, columnspan=2, sticky=S+E+W)

        # ------------ Nakonfigurovani INFO panelu
        info.grid_rowconfigure(20, weight=1)
        info.grid_columnconfigure(0, weight=1)

        # nastavim rozpinani 0;0 - canvas
        self.root.grid_columnconfigure(1, minsize=200, weight=1)
        self.root.grid_rowconfigure(0, weight=1)

        # rozpinani info panelu
        self.root.grid_columnconfigure(3, minsize=200, weight=0)

        self.alg.canvas = self.canvas

        # zacneme nultou operaci
        self.on_step_click()

        # vyzdvihnuti okna
        root.lift()
        root.focus_set()
        self.root.call('wm', 'attributes', '.', '-topmost', True)
        self.root.after_idle(self.root.call, 'wm', 'attributes', '.', '-topmost', False)

    def update_steps(self):
        codes = self.alg.step_codes
        step_index = codes.index(self.alg.step)

        #print codes, step_index, self.alg.step

        last_big_step = None

        for index, label in enumerate(self.labels_steps):
            # Velky krok, obsahujici nekolik podkroku
            if not '-' in codes[index]:
                last_big_step = codes[index]
                font = self.FONT_STEP
            else:
                font = self.FONT_SUB_STEP

            label['font'] = font

            if index < step_index and not self.alg.step.startswith(last_big_step):
                label['fg'] = 'dark green'
            elif index == step_index:
                label['fg'] = 'black'
                label['font'] += ' bold'
            else:
                label['fg'] = '#777'

    def update_step_gui(self, step, result):
        """
        Aktualizace (reakce) GUI na provedeny krok.
        """
        self.step_info['text'] = self.alg.step_info

        # inicializace algoritmu
        if step == 'init':
            if result is None:
                return
            for p in result:
                item = self.canvas.create_oval(p.x-3, p.y-3, p.x+3, p.y+3,
                                               fill='red', outline='white', tags='event')
                # ulozim si cislo prvku do bodu udalosti, at to pak muzu dohledat
                p.tag = item
                self.canvas_labels[item] = 'Bod události p=[{0};{1}]'.format(int(p.x), int(p.y))

        if step == 'event':
            # ten predchozi
            self.canvas_delete_event()

            # je to udalost?
            if result is not None:
                event = result
                # vytvorim objekt zametaci primky
                self.canvas_create_sweepline(event.y)
                self.canvas.itemconfig(event.tag, fill='green')
            else:
                self.canvas_delete_sweepline()

        if step == 'event-sets':
            event = self.alg.event
            # pokud to je prusecik, prebarvim ho na bilou, pokud ne, necham to tak,
            # ale zaridim jeho odstraneni. Result vraci L(p) u C(p) u U(p)
            if len(result) > 1:
                self.canvas.itemconfig(event.tag, fill='white')
                self.canvas_labels[event.tag] = "Průsečík úseček " + ", ".join([s.name for s in result])
            else:
                self._event_to_delete = event.tag

        if step == 'event-findnew':
            for p in result:
                item = self.canvas.create_oval(float(p.x)-3, float(p.y)-3, float(p.x)+3, float(p.y)+3,
                                               fill='yellow', outline='white', tags='event')
                # ulozim si cislo prvku do bodu udalosti, at to pak muzu dohledat
                p.tag = item
                self.canvas_labels[item] = 'Bod události p=[{0};{1}]'.format(int(p.x), int(p.y))

        if step == 'end':
            pass

    def canvas_create_sweepline(self, y):
        w = self.canvas.winfo_width()
        items = self.canvas.find_withtag('sweepline')
        y = float(y)
        # aktualizuji data u stavajici
        if len(items):
            l = items[0]
            self.canvas.coords(l, 0, y, w, y)
        else:
            l = self.canvas.create_line(0, y, w, y, fill='white', dash=(3, 4), tags='sweepline')
            self.canvas_labels[l] = "Zametací přímka"

    def canvas_delete_sweepline(self):
        # nemusi byt zadana ani jedna usecka a tohle muze byt prazdne
        items = self.canvas.find_withtag('sweepline')
        if len(items):
            sweepline = items[0]
            self.canvas.delete(sweepline)
            del self.canvas_labels[sweepline]

    def canvas_delete_event(self):
        if self._event_to_delete is not None:
            self.canvas.delete(self._event_to_delete)
            self._event_to_delete = None

    def on_canvas_btn1_click(self, event):
        if self.alg.step == 'input':
            self._new_segment = self.canvas.create_line(event.x, event.y, event.x, event.y,
                                                        fill='white', width=1)

    def on_canvas_btn1_motion(self, event):
        self.status_pos['text'] = "X={0}  Y={1}".format(event.x, event.y)

        if self.alg.step == 'input':
            x1, y1, x2, y2 = self.canvas.coords(self._new_segment)
            #w, h = self.canvas.winfo_width(), self.canvas.winfo_height()
            #x = min(max(0, event.x), w)
            #y = min(max(0, event.y), h)
            self.canvas.coords(self._new_segment, x1, y1, event.x, event.y)

    def on_canvas_btn1_release(self, event):
        if self.alg.step == 'input':
            # pro zichr
            if self._new_segment is None:
                return
            x1, y1, x2, y2 = self.canvas.coords(self._new_segment)
            # bodova usecka? takove nechceme
            if x1 == x2 and y1 == y2:
                self.canvas.delete(self._new_segment)
                self._new_segment = None
                return

            seg = Segment(Point(x1, y1), Point(x2, y2))
            # tato fce je nezbytna, + nastavi u usecky nazev (seg.name)
            self.alg.add_segment(seg)
            self.canvas.itemconfig(self._new_segment, tags=('segment', seg.name))
            self.canvas_labels[self._new_segment] = 'Úsečka {0}'.format(seg.name)

    def on_canvas_motion(self, event):
        """
        Obsluha udalosti pohybu kurzoru mysi.
        """
        self.status_pos['text'] = "X={0}  Y={1}".format(event.x, event.y)

        items = self.canvas.find_closest(event.x, event.y, 1)

        if len(items) and items[0] != self._item_closest:
            item = items[0]
            # predchozi polozka neni prazdna -> je treba obnovit puvodni konfiguraci
            if self._item_closest is not None:
                self.canvas.itemconfig(self._item_closest, **self._item_closest_config)
            self._item_closest = item
            self._item_closest_config = {'width': self.canvas.itemcget(item, 'width')}
            self.canvas.itemconfig(item, width=2)
            # upgradujeme status objektu
            self.status_item['text'] = self.canvas_labels[item] if item in self.canvas_labels else ''

    def on_canvas_leave(self, event):
        """
        Kurzor opustil platno.
        """
        self.status_pos['text'] = ''
        self.status_item['text'] = ''

        if self._item_closest is not None:
            self.canvas.itemconfig(self._item_closest, **self._item_closest_config)
            self._item_closest = None
            self._item_closest_config = None

    def on_canvas_enter(self, event):
        """
        Kurzor mysi na prostoru platna.
        """
        self.status_pos['text'] = "X={0}  Y={1}".format(event.x, event.y)

    def on_all_click(self):
        """
        Kliknuti na tlacitko "Vse"
        """
        codes = self.alg.step_codes
        curr_index = codes.index(self.alg.step)
        # budeme kroky vykonavat do te doby, nez stavajici predstihne nasledujici
        while True:
            self.on_step_click()
            if codes.index(self.alg.step) > curr_index and '-' not in self.alg.step:
                break

    def on_step_click(self):
        """
        Kliknuti na tlacitko "Dalsi"
        """
        result = self.alg.do_step()
        self.update_steps()
        self.btn_step['state'] = NORMAL if self.alg.can_step else DISABLED
        self.btn_next['state'] = NORMAL if self.alg.can_step else DISABLED
        self.update_step_gui(self.alg.step, result)
        self.root.update_idletasks()


# Inicializace Tk a GUI
tk = Tk()
win = SweepLineGUI(tk)
mainloop()